import cga_py as ga
import numpy as np


def test_arr_to_quat():
    x = np.random.rand(4)
    np.testing.assert_array_almost_equal(
        ga.arr_to_quat(x).coeff,
        (x[0] + x[1] * ga.q_i + x[2] * ga.q_j + x[3] * ga.q_k).coeff,
    )


def test_quat_to_arr():
    x = np.random.rand(4)
    np.testing.assert_array_almost_equal(x, ga.quat_to_arr(ga.arr_to_quat(x)))


def test_rotors():
    r = ga.rand_rotor()
    q0, q1, q2, q3 = ga.rotor_to_quat(r)
    a = ga.quat_to_rotor(q0, q1, q2, q3)
    np.testing.assert_array_almost_equal(r.coeff, a.coeff)
    q0 = ga.quat_to_arr(q0)
    q1 = ga.quat_to_arr(q1)
    q2 = ga.quat_to_arr(q2)
    q3 = ga.quat_to_arr(q3)
    a = ga.quat_to_rotor(q0, q1, q2, q3)
    np.testing.assert_array_almost_equal(r.coeff, a.coeff)


def test_vectorial():
    q = np.random.rand(4)
    p1 = ga.vectorial(q)
    np.testing.assert_array_almost_equal(p1, q[1:-1])
    q2 = q
    q2[0] = 0
    np.testing.assert_array_almost_equal(
        ga.vectorial(ga.arr_to_quat(q)).coeff, ga.arr_to_quat(q2).coeff
    )


def test_scalar():
    q = np.random.rand(4)
    np.testing.assert_array_almost_equal(
        ga.scalar(ga.arr_to_quat(q)).coeff, ga.arr_to_quat([q[0], 0, 0, 0]).coeff
    )
    np.testing.assert_almost_equal(q[0], ga.scalar(q))



from sympy.combinatorics.partitions import Partition
from sympy.combinatorics.permutations import Permutation
import itertools as it

if __name__ == "__main__":
    file = open("../cga_py/permutations.py", "w")
    # generate File header
    file.write("from multivector import *\n\n")

    base2 = [
        "e_12",
        "e_13",
        "e_1i",
        "e_1o",
        "e_23",
        "e_2i",
        "e_2o",
        "e_3i",
        "e_3o",
        "e_io",
    ]
    base3 = [
        "e_123",
        "e_12i",
        "e_12o",
        "e_13i",
        "e_13o",
        "e_1io",
        "e_23i",
        "e_23o",
        "e_2io",
        "e_3io",
    ]
    base4 = ["e_123i", "e_123o", "e_12io", "e_13io", "e_23io"]
    base5 = ["e_123io"]
    perm2 = list(it.permutations([0, 1]))
    perm3 = list(it.permutations([0, 1, 2]))
    perm4 = list(it.permutations([0, 1, 2, 3]))
    perm5 = list(it.permutations([0, 1, 2, 3, 4]))

    def index_permutator(string, permut):
        """TODO: Docstring for index_permutator.
        Returns: TODO

        """
        indices = string.split("_")
        index_num = len(indices[1])
        permutated = indices[0] + "_"
        for i in permut:
            permutated += indices[1][i]
        return permutated

    file.write("# Permutations of order 2 elements\n")
    for x in perm2:
        sig = Permutation(x).signature()
        if int(sig) == 1:
            for base in base2:
                if base == base2[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base + "; ")
        else:
            for base in base2:
                if base == base2[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base + "; ")
        file.write("\n")
    file.write("\n\n")

    file.write("# Permutations of order 3 elements\n")
    for x in perm3:
        sig = Permutation(x).signature()
        if int(sig) == 1:
            for base in base3:
                if base == base3[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base + "; ")
        else:
            for base in base3:
                if base == base3[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base + "; ")
        file.write("\n")
    file.write("\n\n")

    file.write("# Permutations of order 4 elements\n")
    for x in perm4:
        sig = Permutation(x).signature()
        if int(sig) == 1:
            for base in base4:
                if base == base4[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base + "; ")
        else:
            for base in base4:
                if base == base4[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base + "; ")
        file.write("\n")
    file.write("\n\n")

    file.write("# Permutations of order 5 elements\n")
    for x in perm5:
        sig = Permutation(x).signature()
        if int(sig) == 1:
            for base in base5:
                if base == base5[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " =  " + base + "; ")
        else:
            for base in base5:
                if base == base5[-1]:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base)
                else:
                    new = index_permutator(base, x)
                    file.write(new + " = -" + base + "; ")
        file.write("\n")



from cga_py import *
import numpy as np
from scipy.optimize import root as solve


a = e_12 + 3 * e_1o
b = e_1 + e_12
s1 = sphere([0, 0, 0], 1)
s2 = sphere([1, 0, 0], 1)

q0 = arr_to_quat([1, 2, 3, 4])

point_p_act(np.array([[1, 1, 1], [-1, -1, -1]]), 4, [1, iso_scale([0, 0, 0])])



from .multivector import cga_object
from .base_objects import q_i, q_j, q_k, eps_1, eps_2, eps_3
import numpy as np


def arr_to_quat(arr):
    """Converts an array to quaternion with coefficients given in array

    Parameters
    ----------
    arr : array_like
        array giving the coefficients of [1,i,j,k] in this order.

    Returns
    -------
    cga_object
        Quaternion given by the coefficients in `arr` represented in terms of
        CGA.

    """
    return arr[0] + q_i * arr[1] + q_j * arr[2] + q_k * arr[3]


def quat_to_arr(q):
    """Coefficients of quaternion `q` given in terms of CGA.

    Parameters
    ----------
    q : cga_object
        Quaternion of which coefficients are to be calculated.

    Returns
    -------
    array
        Coefficients of quaternion `q` given in order of [1,i,j,k].

    """
    return np.array([q.coeff[0], -q.coeff[10], q.coeff[7], -q.coeff[6]])


def rotor_to_quat(rot):
    """Convert Rotor to its four quaternion representation

    Parameters
    ----------
    rot : cga_object
        Returns: list of quaternions

    Returns
    -------
    tuple
        Tuple of quaternions that represent the rotor in terms of the four
        quaternion decomposition.

    Notes
    -----
    Rotors in CGA can be represented through four quaternions in the form
    q0 + eps_1*q1 + eps_2*q2 + eps_3*q3.
    These quaternions can be directly extracted from the CGA representation of
    the rotor.

    """
    coeff = rot.get_even()
    q0 = (
        (coeff[0] - coeff[10])
        + q_i * (coeff[15] - coeff[5])
        + q_j * (coeff[2] - coeff[14])
        + q_k * (coeff[13] - coeff[1])
    )
    q1 = coeff[11] + q_i * coeff[3] + q_j * coeff[6] + q_k * coeff[8]
    q2 = coeff[12] + q_i * coeff[4] + q_j * coeff[7] + q_k * coeff[9]
    q3 = coeff[10] + q_i * (-coeff[15]) + q_j * coeff[14] + q_k * (-coeff[13])
    return q0, q1, q2, q3


def quat_to_rotor(q0, q1, q2, q3):
    """Compute rotor given by the four quaternions q0,q1,q2,q3

    Parameters
    ----------
    q0 : array_like
        Quaternion given in list-form
    q1 : array_like
        Quaternion given in list-form
    q2 : array_like
        Quaternion given in list-form
    q3 : array_like
        Quaternion given in list-form

    Returns
    -------
    cga_object
        Rotor defined by the four quaternions.

    Notes
    -----
    Rotors can be represented by four quaternionsq0 q1,q2,q3 as lists so that
    h = q0 + eps_1*q1 + eps_2*q2 + eps_3*q3


    """
    if not (
        isinstance(q0, cga_object)
        and isinstance(q1, cga_object)
        and isinstance(q2, cga_object)
        and isinstance(q0, cga_object)
    ):
        p0 = q0[0] + q_i * q0[1] + q_j * q0[2] + q_k * q0[3]
        p1 = q1[0] + q_i * q1[1] + q_j * q1[2] + q_k * q1[3]
        p2 = q2[0] + q_i * q2[1] + q_j * q2[2] + q_k * q2[3]
        p3 = q3[0] + q_i * q3[1] + q_j * q3[2] + q_k * q3[3]
    else:
        p0 = q0
        p1 = q1
        p2 = q2
        p3 = q3

    return p0 + eps_1 * p1 + eps_2 * p2 + eps_3 * p3


def vectorial(quat):
    """Returns vectorial part of quaternion

    Parameters
    ----------
    quat : cga_object or array_like
        Quaternion of which to to extract the vectorial component.

    Returns
    -------
    cga_object or array_like
        Vectorial part of quaternion in same representation as input datatype.

    """
    if isinstance(quat, cga_object):
        return 1 / 2 * (quat - (~quat))
    else:
        return quat[1:-1]


def scalar(quat):
    """Returns scalar part of quaternion

    Parameters
    ----------
    quat : cga_object or array_like
        Quaternion of which to extract the scalar component.

    Returns
    -------
    cga_object or array_like
        Vectorial part of quaternion in same representation as input datatype.

    """
    if isinstance(quat, cga_object):
        return quat - 1 / 2 * (quat - (~quat))
    else:
        return quat[0]



from .multivector import cga_object
from .base_objects import e_1, e_2, e_3, e_i, e_o
import numpy as np
import cmath as cm


def point(point):
    """Generate Point in (x,y,z)

    Parameters
    ----------
    point : array_like
        Cartesian coordinates of point given as [x, y, z]

    Returns
    -------
    cga_object
        CGA representation of point with coordinates `point`

    Notes
    -----
    In CGA all objects can be interpreted as spheres given by:
        x*e_1 + y*e_2 + z*e_3 + 1/2*(x^2 + y^2 + z^2 - r^2)*e_i + e_0
    Thus a point can be seen as a sphere with radius r = 0 with center
    [x, y, z].
    So a point is represented as:
        x*e_1 + y*e_2 + z*e_3 + 1/2*(x^2 + y^2 + z^2)*e_i + e_0

    """
    if len(point) != 3:
        raise ValueError("Point must have three coordinates.")
    return (
        point[0] * e_1
        + point[1] * e_2
        + point[2] * e_3
        + 1 / 2 * (point[0] ** 2 + point[1] ** 2 + point[2] ** 2) * e_i
        + e_o
    )


def normalize_point(point):
    """Normalize CGA representation of point

    Parameters
    ----------
    point : cga_object
        Point in conformal representation.

    Returns
    -------
    cga_object
        `point` given in its normalized CGA representation.

    """
    num_max = max(abs(point.coeff[1:5]))
    if not np.isclose(point.coeff[0] / num_max, 0):
        raise ValueError(str(point) + " is not a cga representation of a point")

    if not np.allclose(point.coeff[6:-1] / num_max, np.zeros_like(point.coeff[6:-1])):
        raise ValueError(str(point) + " is not a cga representation of a point")

    return cga_object(1 / point.coeff[5] * point.coeff)


def point_to_cartesian(point):
    """Convert CGA representation of a point into cartesian coordinates.

    Parameters
    ----------
    point : cga_object
        Point to be converted into cartesian system.

    Returns
    -------
    array_like
        Cartesian coordinates of `point`.

    """
    return np.array(normalize_point(point).coeff[1:4], dtype=complex)


def sphere(center, radius):
    """Generate conformal representation of a sphere in `center` with `radius`

    Parameters
    ----------
    center : array_like
        Center of sphere as [x,y,z].
    radius : float
        Radius of sphere.

    Returns
    -------
    cga_object
        CGA representation of Sphere with center in `center` and radius
        `radius`.

    Notes
    -----
    In CGA all objects can be interpreted as spheres given by:
        x*e_1 + y*e_2 + z*e_3 + 1/2*(x^2 + y^2 + z^2 - r^2)*e_i + e_0

    """
    if len(center) != 3:
        raise ValueError("Center must have three coordinates.")
    return (
        center[0] * e_1
        + center[1] * e_2
        + center[2] * e_3
        + (1 / 2)
        * (center[0] ** 2 + center[1] ** 2 + center[2] ** 2 - radius**2)
        * e_i
        + e_o
    )


def normalize_sphere(sphere):
    """Normalize CGA representation of sphere.

    Parameters
    ----------
    sphere : cga_object
        CGA representation of a sphere.

    Returns
    -------
    cga_object
        CGA representation of the sphere normalized to e0.


    """
    num_max = max(abs(sphere.coeff[1:5]))
    if not np.isclose(sphere.coeff[0] / num_max, 0):
        raise ValueError("Object is not a cga representation of a sphere")

    if not np.allclose(sphere.coeff[6:-1] / num_max, np.zeros_like(sphere.coeff[6:-1])):
        raise ValueError("Object is not a cga representation of a sphere")
    return cga_object(1 / sphere.coeff[5] * sphere.coeff)


def sphere_to_cartesian(sphere):
    """Convert sphere in CGA representation to center and radius.

    Parameters
    ----------
    sphere : cga_object
        Sphere to be converted to cartesian.

    Returns
    -------
    (array_like, float)
        Center of sphere given as [x, y, z] and radius of sphere.

    """
    norm_sphere = normalize_sphere(sphere)
    x = norm_sphere.coeff[1]
    y = norm_sphere.coeff[2]
    z = norm_sphere.coeff[3]
    r = x**2 + y**2 + z**2 - 2 * norm_sphere.coeff[4]
    return np.array([x, y, z]), cm.sqrt(r)


def plane(normal, distance):
    """Generate conformal representation of s plane with normalvector normal and
    distance to origin distance

    Parameters
    ----------
    normal : nd.array
        Normal vector given as [x,y,z]
    distance : float
        Distance to origin
        Returns: cga_object

    Returns
    -------

    """
    if len(normal) != 3:
        raise ValueError("Normal must have three coordinates.")
    c = normal / np.sqrt(normal[0] ** 2 + normal[1] ** 2 + normal[2] ** 2)
    return c[0] * e_1 + c[1] * e_2 + c[2] * e_3 + distance * e_i


def normalize_plane(plane):
    """Normalize CGA representation of a plane

    Parameters
    ----------
    plane : cga_object
        Plane to be normalized
        Returns: (cga_object)

    Returns
    -------

    """
    num_max = max(abs(plane.coeff[1:5]))
    if not np.isclose(plane.coeff[0] / num_max, 0):
        raise ValueError("Object is not a cga representation of a plane")

    if not np.allclose(plane.coeff[6:-1] / num_max, np.zeros_like(plane.coeff[6:-1])):
        raise ValueError("Object is not a cga representation of a plane")

    norm = 1 / np.sqrt(plane.coeff[1] ** 2 + plane.coeff[2] ** 2 + plane.coeff[3] ** 2)
    return cga_object(norm * plane.coeff)


def plane_to_cartesian(plane):
    """Convert plane given in CGA to representation ad normal vector and
    distance from origin.

    Parameters
    ----------
    plane : cga_object
        plane to be converted
        Returns: (nd.array, float) normal unit-vector, distance from origin

    Returns
    -------

    """
    c = normalize_plane(plane).coeff
    return np.array([c[1], c[2], c[3]]), c[4]



from .base_objects import e_1, e_2, e_3, e_i
from .multivector import cga_object
from .geom_generators import point_to_cartesian, point
from .errors import GeometryError


def iso_scale(a):
    """Generates the rotor for an isotropic scaling with center a

    Parameters
    ----------
    a : cga_object, ndarray
        scaling center as point in CGA representation

    Returns
    -------
    cga_object
        rotor corresponding to a uniform scaling with center a

    """
    if not isinstance(a, cga_object):
        a = point(a)
    return a ^ e_i


def transv(a, b):
    """Generates transversion defined by the point a and the orthogonal plane b

    Parameters
    ----------
    a : cga_object, ndarray
        Point a given as cga_object or as list of coordinates
    b : cga_object, ndarray
        Plane given as cga_object or as list of coordinates for its normal-vector
    Returns
    -------
    cga_object
        rotor corresponding to a transversion defined by a and b.

    """
    # Check if a is given in CGA representation or as coordinates
    if isinstance(a, cga_object):
        coord = point_to_cartesian(a)
    else:
        # Extract coordinates and generate the CGA representation
        coord = a
        a = point(coord)
    if isinstance(b, cga_object):
        if (a | b) != cga_object([0]):
            raise GeometryError("Plane b must be orthogonal to point a")
    else:
        b = (
            b[0] * e_1
            + b[1] * e_2
            + b[2] * e_3
            + (coord[0] * b[0] + coord[1] * b[1] + coord[2] * b[2]) * e_i
        )

    return a ^ b



import cga_py as ga
import numpy as np


def test_poly_act():
    a = ga.cga_object(np.random.rand(32))
    x0 = ga.cga_object(np.random.rand(32))
    x1 = ga.cga_object(np.random.rand(32))
    x2 = ga.cga_object(np.random.rand(32))
    x3 = ga.cga_object(np.random.rand(32))
    x4 = ga.cga_object(np.random.rand(32))
    x5 = ga.cga_object(np.random.rand(32))
    t = np.random.rand()
    p = ga.poly_act(t, [x0, x1, x2, x3, x4, x5], a)
    pp = x0 + x1 * t + x2 * (t**2) + x3 * (t**3) + x4 * (t**4) + x5 * (t**5)
    b = ga.act(pp, a)
    np.testing.assert_array_almost_equal(p.coeff, b.coeff)



from .operators import act


def poly_act(param, poly, obj):
    """Let the motion polynomial poly(param) act on obj

    Parameters
    ----------
    param : float
        parameter of polynomial
    poly : ndarray
        polynomial represented by coefficients in list form starting at degree
        0 and increasing with array length
    obj : cga_object
        Object on which to let the rotor polynomial act

    Returns
    -------
    cga_object

    """
    return act(sum([poly[i] * (param**i) for i in range(len(poly))]), obj)



from .multivector import cga_object
import numpy as np


def act(rotator, obj):
    """Function to let rotator act on obj

    Parameters
    ----------
    rotator : cga_object
        rotator to apply to obj
    obj : cga_object
        Object to be acted upon


    Returns
    -------
    cga_object
        Rotated object

    """
    return rotator * obj * (~rotator)


def com(a, b):
    """commutator Product of a and b

    Parameters
    ----------
    a : cga_object
    b : cga_object

    Returns
    -------
    cga_object

    """
    return (1 / 2) * (a * b - b * a)


def anti_com(a, b):
    """anti-commutator Product of a and b

    Parameters
    ----------
    a : cga_object
    b : cga_object

    Returns
    -------
    cga_object

    """
    return (1 / 2) * (a * b + b * a)


def n_grade(obj, grade):
    """Returns parts of obj with grade grade

    Parameters
    ----------
    obj : cga_object
        object of which to extract grade-graded parts
    grade : int
        grade of parts to be extracted

    Returns
    -------
    cga_object
        grade-graded parts of obj

    """
    vec = np.zeros_like(obj.coeff)
    if grade == 0:
        ind = 0
        vec[ind] = obj.coeff[ind]
        return cga_object(vec)
    elif grade == 1:
        ind = range(1, 6)
        vec[ind] = obj.coeff[ind]
        return cga_object(vec)
    elif grade == 2:
        ind = range(6, 16)
        vec[ind] = obj.coeff[ind]
        return cga_object(vec)
    elif grade == 3:
        ind = range(16, 26)
        vec[ind] = obj.coeff[ind]
        return cga_object(vec)
    elif grade == 4:
        ind = range(26, 31)
        vec[ind] = obj.coeff[ind]
        return cga_object(vec)
    else:
        ind = 31
        vec[ind] = obj.coeff[ind]
        return cga_object(vec)


def r_norm(obj):
    """Calculates the right norm in CGA

    Parameters
    ----------
    obj :
        cga_object

    Returns
    -------
    cga_object


    """
    return obj * (~obj)


def l_norm(obj):
    """Calculates the left norm in CGA

    Parameters
    ----------
    obj :
        cga_object

    Returns
    -------
    cga_object


    """
    return (~obj) * obj



#!/usr/bin/env python

from cga_py import *
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets
import pyqtgraph.opengl as gl
import pyqtgraph.parametertree.parameterTypes as pTypes
from pyqtgraph.parametertree import Parameter, ParameterTree
from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea
import numpy as np


########################################
# View generation
########################################
# Generate the App
app = pg.mkQApp("CGA-Py Visualizer")
# Make the window in which the app is running
win = QtWidgets.QMainWindow()
# Make the are in which docking is posible and then set it as the central widget
area = DockArea()
win.setCentralWidget(area)

####################
# Create the docks
####################

# Make the two docks. D1 has minimal size and d2 whatever was in the example
d1 = Dock("Timeline", size=(1, 1))
d2 = Dock("Visualization", size=(100, 400))
d3 = Dock("Parameters", size=(50, 400))

# Add the docks to the docking area
area.addDock(d2, "top")
area.addDock(d3, "left")
area.addDock(d1, "bottom")


####################
# Timeline Dock
####################
# make a parameter collection
time_parameters = Parameter.create(name="params", type="group")

# Add a slider
p_t_slider = time_parameters.addChild(pTypes.SliderParameter(name="Time Slider"))
# change settings
p_t_slider.setLimits([-np.pi, np.pi])
p_t_slider.setOpts(step=np.pi / 300)
p_t_slider.setValue(np.pi / 2 + 0.001)
p_t_slider.setDefault(np.pi / 2 + 0.001)
# Add a slider
p_t_value = time_parameters.addChild(
    pTypes.SimpleParameter(type="str", name="Time Input")
)
p_t_value.setValue(p_t_slider.value())
p_t_value.setDefault(p_t_slider.value())


# What happens at update of parameter
def update_t_slide():
    # p_t_value.setValue(p_t_slider.value())
    full_time_update()


def update_t_val():
    p_t_slider.setValue(eval(p_t_value.value()))
    full_time_update()


# Create a parameter tree for displaying the slider
time_paramtree = ParameterTree()
time_paramtree.setParameters(time_parameters, showTop=False)

####################
# Parameters dock
####################

general_params = Parameter.create(name="params", type="group")
p_update_all = general_params.addChild(pTypes.ActionParameter(name="Update All"))
p_execute_code = general_params.addChild(pTypes.ActionParameter(name="Execute Code"))
p_shell_interface = general_params.addChild(pTypes.TextParameter(name="Custom Code"))
p_shell_interface.setValue(
    "#This Field can execute custom code. Variables must be declared with the global keyword!"
)

axes_tree = general_params.addChild(Parameter.create(name="Axes", type="group"))
p_show_axes = axes_tree.addChild(pTypes.SimpleParameter(name="Show Axes", type="bool"))
p_show_axes.setValue(True)
p_show_axes.setDefault(True)
p_axes_size = axes_tree.addChild(
    pTypes.SimpleParameter(name="Axes Lengths", type="str")
)
p_axes_size.setValue("[3,3,3]")
p_axes_size.setDefault("[3,3,3]")


# Polynomial input
poly_tree = general_params.addChild(Parameter.create(name="Polynomial", type="group"))

p_update_coeff = poly_tree.addChild(pTypes.ActionParameter(name="Update Coefficients"))


p_use_main_coeff = poly_tree.addChild(
    pTypes.SimpleParameter(name="Use Main Polynomial", type="bool")
)
p_use_main_coeff.setValue(True)


p_main_poly_coeff = poly_tree.addChild(pTypes.TextParameter(name="Main Coeff"))
p_main_poly_coeff.setValue("[0.5*e_123o - e_123i, e_12 - e_3i + 0.5*e_3o, 1]")
p_main_poly_coeff.setDefault("[0.5*e_123o - e_123i, e_12 - e_3i + 0.5*e_3o, 1]")

x = 1
y = 1
z = 100

p_first_poly_coeff = poly_tree.addChild(pTypes.TextParameter(name="First Coeff"))
p_first_poly_coeff.setValue(
    "[-1/2*(2*x**4*e_3i-x**4*e_3o+4*x**2*y**2*e_3i-2*x**2*y**2*e_3o+4*x**2*z**2*e_3i-2*x**2*z**2*e_3o+2*y**4*e_3i-y**4*e_3o+4*y**2*z**2*e_3i-2*y**2*z**2*e_3o+2*z**4*e_3i-z**4*e_3o-8*x**3*e_13+8*x**3*e_2i-4*x**3*e_2o-8*x**2*y*e_1i+4*x**2*y*e_1o-8*x**2*y*e_23-8*x*y**2*e_13+8*x*y**2*e_2i-4*x*y**2*e_2o-8*x*z**2*e_13+8*x*z**2*e_2i-4*x*z**2*e_2o-8*y**3*e_1i+4*y**3*e_1o-8*y**3*e_23-8*y*z**2*e_1i+4*y*z**2*e_1o-8*y*z**2*e_23-32*x**2*e_12-16*x**2*e_3i+8*x**2*e_3o+32*x*z*e_1i-16*x*z*e_1o+32*x*z*e_23-32*y**2*e_12-16*y**2*e_3i+8*y**2*e_3o-32*y*z*e_13+32*y*z*e_2i-16*y*z*e_2o+16*z**2*e_3i-8*z**2*e_3o+32*x*e_13-32*x*e_2i+16*x*e_2o+32*y*e_1i-16*y*e_1o+32*y*e_23+32*e_3i-16*e_3o)/(x**2+y**2+z**2+4)**2, 1]"
)
p_first_poly_coeff.setDefault(
    "[-1/2*(2*x**4*e_3i-x**4*e_3o+4*x**2*y**2*e_3i-2*x**2*y**2*e_3o+4*x**2*z**2*e_3i-2*x**2*z**2*e_3o+2*y**4*e_3i-y**4*e_3o+4*y**2*z**2*e_3i-2*y**2*z**2*e_3o+2*z**4*e_3i-z**4*e_3o-8*x**3*e_13+8*x**3*e_2i-4*x**3*e_2o-8*x**2*y*e_1i+4*x**2*y*e_1o-8*x**2*y*e_23-8*x*y**2*e_13+8*x*y**2*e_2i-4*x*y**2*e_2o-8*x*z**2*e_13+8*x*z**2*e_2i-4*x*z**2*e_2o-8*y**3*e_1i+4*y**3*e_1o-8*y**3*e_23-8*y*z**2*e_1i+4*y*z**2*e_1o-8*y*z**2*e_23-32*x**2*e_12-16*x**2*e_3i+8*x**2*e_3o+32*x*z*e_1i-16*x*z*e_1o+32*x*z*e_23-32*y**2*e_12-16*y**2*e_3i+8*y**2*e_3o-32*y*z*e_13+32*y*z*e_2i-16*y*z*e_2o+16*z**2*e_3i-8*z**2*e_3o+32*x*e_13-32*x*e_2i+16*x*e_2o+32*y*e_1i-16*y*e_1o+32*y*e_23+32*e_3i-16*e_3o)/(x**2+y**2+z**2+4)**2, 1]"
)

p_second_poly_coeff = poly_tree.addChild(pTypes.TextParameter(name="Second Coeff"))
p_second_poly_coeff.setValue(
    "[-16/(x**2+y**2+z**2+4)**2*e_12*x**2-16/(x**2+y**2+z**2+4)**2*e_12*y**2-4/(x**2+y**2+z**2+4)**2*e_13*x**3+16/(x**2+y**2+z**2+4)**2*e_13*x-4/(x**2+y**2+z**2+4)**2*e_1i*y**3+16/(x**2+y**2+z**2+4)**2*e_1i*y+2/(x**2+y**2+z**2+4)**2*e_1o*y**3-8/(x**2+y**2+z**2+4)**2*e_1o*y-4/(x**2+y**2+z**2+4)**2*e_23*y**3+16/(x**2+y**2+z**2+4)**2*e_23*y+4/(x**2+y**2+z**2+4)**2*e_2i*x**3-16/(x**2+y**2+z**2+4)**2*e_2i*x-2/(x**2+y**2+z**2+4)**2*e_2o*x**3+8/(x**2+y**2+z**2+4)**2*e_2o*x+16/(x**2+y**2+z**2+4)**2*e_3i+1/(x**2+y**2+z**2+4)**2*e_3i*x**4+1/(x**2+y**2+z**2+4)**2*e_3i*y**4+1/(x**2+y**2+z**2+4)**2*e_3i*z**4-8/(x**2+y**2+z**2+4)**2*e_3i*x**2-8/(x**2+y**2+z**2+4)**2*e_3i*y**2+8/(x**2+y**2+z**2+4)**2*e_3i*z**2-8/(x**2+y**2+z**2+4)**2*e_3o-1/2/(x**2+y**2+z**2+4)**2*e_3o*x**4-1/2/(x**2+y**2+z**2+4)**2*e_3o*y**4-1/2/(x**2+y**2+z**2+4)**2*e_3o*z**4+4/(x**2+y**2+z**2+4)**2*e_3o*x**2+4/(x**2+y**2+z**2+4)**2*e_3o*y**2-4/(x**2+y**2+z**2+4)**2*e_3o*z**2+2/(x**2+y**2+z**2+4)**2*e_3i*x**2*y**2+2/(x**2+y**2+z**2+4)**2*e_3i*x**2*z**2+2/(x**2+y**2+z**2+4)**2*e_3i*z**2*y**2-1/(x**2+y**2+z**2+4)**2*e_3o*x**2*y**2-1/(x**2+y**2+z**2+4)**2*e_3o*x**2*z**2-1/(x**2+y**2+z**2+4)**2*e_3o*z**2*y**2-4/(x**2+y**2+z**2+4)**2*e_13*x*y**2-4/(x**2+y**2+z**2+4)**2*e_13*x*z**2-16/(x**2+y**2+z**2+4)**2*e_13*y*z-4/(x**2+y**2+z**2+4)**2*e_1i*x**2*y-4/(x**2+y**2+z**2+4)**2*e_1i*y*z**2+16/(x**2+y**2+z**2+4)**2*e_1i*x*z+2/(x**2+y**2+z**2+4)**2*e_1o*x**2*y+2/(x**2+y**2+z**2+4)**2*e_1o*y*z**2-8/(x**2+y**2+z**2+4)**2*e_1o*x*z-4/(x**2+y**2+z**2+4)**2*e_23*x**2*y-4/(x**2+y**2+z**2+4)**2*e_23*y*z**2+16/(x**2+y**2+z**2+4)**2*e_23*x*z+4/(x**2+y**2+z**2+4)**2*e_2i*x*y**2+4/(x**2+y**2+z**2+4)**2*e_2i*x*z**2+16/(x**2+y**2+z**2+4)**2*e_2i*y*z-2/(x**2+y**2+z**2+4)**2*e_2o*x*y**2-2/(x**2+y**2+z**2+4)**2*e_2o*x*z**2-8/(x**2+y**2+z**2+4)**2*e_2o*y*z+e_12-e_3i+1/2*e_3o, 1]"
)


####################
# Trajectory Parameters
####################
traj_tree = general_params.addChild(Parameter.create(name="Trajectories", type="group"))

p_traj_points_update = traj_tree.addChild(
    pTypes.ActionParameter(name="Update Trajectories")
)

p_display_tajectory = traj_tree.addChild(
    pTypes.SimpleParameter(name="Display Trajectories", type="bool")
)
p_display_tajectory.setValue(False)

p_traj_points = traj_tree.addChild(pTypes.TextParameter(name="Trajectory Points"))
p_traj_points.setValue("[[np.sin(t),np.cos(t),0] for t in np.linspace(0,2*np.pi,20)]")
p_traj_points.setDefault("[[np.sin(t),np.cos(t),0] for t in np.linspace(0,2*np.pi,20)]")

p_traj_subds = traj_tree.addChild(
    pTypes.SimpleParameter(type="int", name="Trajectory Subdivisions")
)
p_traj_subds.setValue(100)
p_traj_subds.setDefault(100)

p_traj_width = traj_tree.addChild(
    pTypes.SimpleParameter(type="float", name="Trajectory width")
)
p_traj_width.setValue(2)
p_traj_width.setDefault(2)


p_traj_c_map = traj_tree.addChild(pTypes.ColorMapParameter(name="Trajectory Colors"))


####################
# Cube Parameters
####################
cube_tree = general_params.addChild(Parameter.create(name="Cube", type="group"))
p_display_cube = cube_tree.addChild(
    pTypes.SimpleParameter(name="Display Cube", type="bool")
)
p_display_cube.setValue(True)
p_cube_subds = cube_tree.addChild(
    pTypes.SimpleParameter(name="Subdivisions", type="str")
)
p_cube_subds.setOpts(step=1)
p_cube_subds.setValue("[10, 10, 10]")
p_cube_subds.setDefault("[10, 10, 10]")
p_cube_center = cube_tree.addChild(pTypes.SimpleParameter(name="Center", type="str"))
p_cube_center.setOpts(step=1)
p_cube_center.setValue("[0, 0, 0]")
p_cube_center.setDefault("[0, 0, 0]")
p_cube_length = cube_tree.addChild(pTypes.SimpleParameter(name="Lengths", type="str"))
p_cube_length.setOpts(step=1)
p_cube_length.setValue("[2, 2, 2]")
p_cube_length.setDefault("[2, 2, 2]")

####################
# Sphere Parameters
####################
sphere_tree = general_params.addChild(Parameter.create(name="Sphere", type="group"))
p_update_spheres = sphere_tree.addChild(pTypes.ActionParameter(name="Update Spheres"))
p_display_sphere = sphere_tree.addChild(
    pTypes.SimpleParameter(name="Display Sphere", type="bool")
)
p_display_sphere.setValue(False)
p_sphere_params = sphere_tree.addChild(pTypes.TextParameter(name="Center"))
p_sphere_params.setValue(
    "[[[1, 0, 0],0.5],[[0, -1, 0],0.5],[[-1, 0, 0],0.5],[[0, 1, 0],0.5]]"
)
p_sphere_params.setDefault(
    "[[[1, 0, 0],0.5],[[0, -1, 0],0.5],[[-1, 0, 0],0.5],[[0, 1, 0],0.5]]"
)
p_sphere_c_map = sphere_tree.addChild(pTypes.ColorMapParameter(name="Sphere Colors"))


# Create a parameter tree for displaying the slider
general_paramtree = ParameterTree()
general_paramtree.setParameters(general_params, showTop=False)

main_poly_coeff = eval(p_main_poly_coeff.value())
first_poly_coeff = eval(p_first_poly_coeff.value())
second_poly_coeff = eval(p_second_poly_coeff.value())
cube_points, cols = point_cube_gen(
    eval(p_cube_center.value()), eval(p_cube_length.value()), eval(p_cube_subds.value())
)
traj_points = eval(p_traj_points.value())
traj_plots = [gl.GLLinePlotItem(), gl.GLLinePlotItem()]
traj_colors = p_traj_c_map.value().getLookupTable(
    nPts=len(traj_plots), mode=pg.ColorMap.QCOLOR
)
unit_sphere = gl.MeshData.sphere(20, 20)
sphere_params = eval(p_sphere_params.value())
sphere_centers = [sphere_params[i][0] for i in range(len(sphere_params))]
sphere_radii = [sphere_params[i][1] for i in range(len(sphere_params))]
view_spheres = [gl.GLMeshItem(meshdata=unit_sphere, color=(0.5, 0.5, 0.5, 1))]
sphere_colors = p_sphere_c_map.value().getLookupTable(
    nPts=len(view_spheres), mode=pg.ColorMap.QCOLOR
)

########################################
# Create the view and setup parameters
########################################

# Create a 3d viewport
view = gl.GLViewWidget()
# add axes for viewing pleasure
axes = gl.GLAxisItem()
axes.setSize(3, 3, 3)
view.addItem(axes)

# Add the Tree and Viewer to their respective docks
d1.addWidget(time_paramtree)
d2.addWidget(view)
d3.addWidget(general_paramtree)

# Plot them as scatter
scatter = gl.GLScatterPlotItem(pos=cube_points, color=cols)


def update_show_axes():
    if p_show_axes.value():
        axes.setVisible(True)
    else:
        axes.setVisible(False)


def update_axes_size():
    scale = eval(p_axes_size.value())
    axes.setSize(scale[0], scale[1], scale[2])


def point_p_act_main_on_points(point_arr):
    return point_p_act(np.tan(p_t_slider.value()), main_poly_coeff, point_arr)


def point_p_act_factorization_on_points(point_arr):
    return point_p_act(
        np.tan(p_t_slider.value()),
        second_poly_coeff,
        point_p_act(np.tan(p_t_slider.value()), first_poly_coeff, point_arr),
    )


def act_main_on_points(param, actpoint):
    return poly_act(np.tan(param), main_poly_coeff, actpoint)


def act_factorization_on_points(param, actpoint):
    return poly_act(
        np.tan(param),
        second_poly_coeff,
        poly_act(np.tan(param), first_poly_coeff, actpoint),
    )


def update_scatter_with_main():
    scatter.setData(
        pos=point_p_act_main_on_points(cube_points),
        color=cols,
    )


def update_scatter_with_factorization():
    scatter.setData(
        pos=point_p_act_factorization_on_points(cube_points),
        color=cols,
    )


def full_time_update():
    if p_display_cube.value():
        if p_use_main_coeff.value():
            update_scatter_with_main()
        else:
            update_scatter_with_factorization()
    if p_display_sphere.value():
        update_spheres()


def update_display_cube():
    if p_display_cube.value():
        scatter.setVisible(True)
    else:
        scatter.setVisible(False)
    full_time_update()


def generatre_trajectory_points(start_point):
    time = np.linspace(0.01, np.pi + 0.01, p_traj_subds.value())
    points = np.empty((len(time), 3))
    if p_use_main_coeff.value():
        for i in range(len(time)):
            points[i] = np.real(
                point_to_cartesian(act_main_on_points(time[i], point(start_point)))
            )
    else:
        for i in range(len(time)):
            points[i] = np.real(
                point_to_cartesian(
                    act_factorization_on_points(time[i], point(start_point))
                )
            )
    return points


def clear_traj_plots():
    for i in range(len(traj_plots)):
        try:
            view.removeItem(traj_plots[i])
        except ValueError:
            pass


def add_traj_plots():
    for plot in traj_plots:
        view.addItem(plot)


def update_trajectories():
    global traj_points
    global traj_plots
    traj_points = eval(p_traj_points.value())
    clear_traj_plots()
    traj_plots = [gl.GLLinePlotItem(width=5) for i in range(len(traj_points))]
    for i in range(len(traj_plots)):
        traj_plots[i].setData(pos=generatre_trajectory_points(traj_points[i]))
    update_display_trajectory()


def update_trajectory_width():
    for i in range(len(traj_plots)):
        traj_plots[i].setData(width=p_traj_width.value())


def update_trajectory_colors():
    global traj_colors
    traj_colors = p_traj_c_map.value().getLookupTable(
        nPts=len(traj_plots), mode=pg.ColorMap.QCOLOR
    )
    for i in range(len(traj_plots)):
        traj_plots[i].setData(color=traj_colors[i])


def update_display_trajectory():
    if p_display_tajectory.value():
        add_traj_plots()
        update_trajectory_width()
        update_trajectory_colors()
    else:
        clear_traj_plots()


def update_cube():
    global cube_points
    global cols
    cube_points, cols = point_cube_gen(
        eval(p_cube_center.value()),
        eval(p_cube_length.value()),
        np.array(eval(p_cube_subds.value())),
    )
    full_update()


def update_coeff_set():
    full_update()


def update_main_poly_coeff():
    global main_poly_coeff
    main_poly_coeff = eval(p_main_poly_coeff.value())
    full_update()


def update_first_poly_coeff():
    global first_poly_coeff
    first_poly_coeff = eval(p_first_poly_coeff.value())
    full_update()


def update_second_poly_coeff():
    global second_poly_coeff
    second_poly_coeff = eval(p_second_poly_coeff.value())
    full_update()


def add_spheres():
    for sphere in view_spheres:
        view.addItem(sphere)


def remove_spheres():
    for sphere in view_spheres:
        try:
            view.removeItem(sphere)
        except ValueError:
            continue


def update_display_sphere():
    if p_display_sphere.value():
        add_spheres()
    else:
        remove_spheres()


def update_spheres():
    global view_spheres
    if p_display_sphere.value():
        remove_spheres()
    view_spheres = [
        gl.GLMeshItem(meshdata=unit_sphere, color=(0.5, 0.5, 0.5, 1))
        for i in range(len(sphere_centers))
    ]
    update_sphere_colors()
    for i in range(len(sphere_centers)):
        if p_use_main_coeff.value():
            new_sphere = act_main_on_points(
                p_t_slider.value(), sphere(sphere_centers[i], sphere_radii[i])
            )
        else:
            new_sphere = act_factorization_on_points(
                p_t_slider.value(), sphere(sphere_centers[i], sphere_radii[i])
            )
        cent, rad = sphere_to_cartesian(new_sphere)
        cent = np.real(cent)
        rad = np.real(rad)
        view_spheres[i].resetTransform()
        view_spheres[i].setMeshData(meshdata=unit_sphere)
        view_spheres[i].translate(cent[0], cent[1], cent[2])
        view_spheres[i].scale(rad, rad, rad)
    if p_display_sphere.value():
        add_spheres()


def update_sphere_params():
    global sphere_params
    global sphere_centers
    global sphere_radii
    sphere_params = eval(p_sphere_params.value())
    sphere_centers = [sphere_params[i][0] for i in range(len(sphere_params))]
    sphere_radii = [sphere_params[i][1] for i in range(len(sphere_params))]
    update_spheres()


def update_sphere_colors():
    global sphere_colors
    sphere_colors = p_sphere_c_map.value().getLookupTable(
        nPts=len(view_spheres), mode=pg.ColorMap.QCOLOR
    )
    for i in range(len(view_spheres)):
        view_spheres[i].setColor(sphere_colors[i])


def execute_custom_code():
    # Here it is NECESSARY for variables to be declared globally
    exec(p_shell_interface.value())


def full_update():
    full_time_update()


def update_all_params():
    update_sphere_params()
    update_main_poly_coeff()
    update_first_poly_coeff()
    update_second_poly_coeff()
    update_trajectories()


view.addItem(scatter)

# Connect the slider change signals to the update functions
p_update_all.sigActivated.connect(update_all_params)

p_t_slider.sigValueChanged.connect(update_t_slide)
p_t_value.sigValueChanged.connect(update_t_val)

p_execute_code.sigActivated.connect(execute_custom_code)

p_show_axes.sigValueChanged.connect(update_show_axes)
p_axes_size.sigValueChanged.connect(update_axes_size)

p_use_main_coeff.sigValueChanged.connect(update_coeff_set)
p_update_coeff.sigActivated.connect(update_main_poly_coeff)
p_update_coeff.sigActivated.connect(update_first_poly_coeff)
p_update_coeff.sigActivated.connect(update_second_poly_coeff)


p_display_cube.sigValueChanged.connect(update_display_cube)
p_cube_subds.sigValueChanged.connect(update_cube)
p_cube_center.sigValueChanged.connect(update_cube)
p_cube_length.sigValueChanged.connect(update_cube)


p_display_tajectory.sigValueChanged.connect(update_display_trajectory)
p_traj_points_update.sigActivated.connect(update_trajectories)
p_traj_width.sigValueChanged.connect(update_trajectory_width)
p_traj_c_map.sigValueChanged.connect(update_trajectory_colors)

p_display_sphere.sigValueChanged.connect(update_display_sphere)
p_update_spheres.sigActivated.connect(update_sphere_params)
p_sphere_c_map.sigValueChanged.connect(update_sphere_colors)


# Update all parameters before first view
update_all_params()
full_update()

# show the window
win.show()


# The typical stuff for standalone execution
if __name__ == "__main__":
    pg.exec()



import cga_py as ga
import numpy as np


def test_geometry_error_nomessage():
    with np.testing.assert_raises(ga.GeometryError):
        raise ga.GeometryError()
    with np.testing.assert_raises(ga.GeometryError):
        raise ga.GeometryError("test")



import numpy as np
from .errors import GeometryError


def study_cond(a, b):
    """Computes Study condition for objects a and b embeded into the CGA

    Parameters
    ----------
    a : cga_object
        First parameter
    b : cga_object

    Returns
    -------
    cga_object
        Study condition of `a` and `b`
    """
    return a * (~b) + b * (~a)


def study_var(r):
    """Compute Study variety equations for even graded elements of CGA

    Parameters
    ----------
    r : cga_object
        even graded cga_object

    Returns
    -------
    ndarray

    """
    b = r.get_even()
    return np.array(
        [
            b[0] * b[11] - b[1] * b[8] + b[2] * b[6] - b[3] * b[5],
            b[0] * b[12] - b[1] * b[9] + b[2] * b[7] - b[4] * b[5],
            b[0] * b[13] - b[1] * b[10] + b[3] * b[7] - b[4] * b[6],
            b[0] * b[14] - b[2] * b[10] + b[3] * b[9] - b[4] * b[8],
            b[0] * b[15] - b[5] * b[10] + b[6] * b[9] - b[7] * b[8],
            -b[1] * b[14] + b[2] * b[13] - b[3] * b[12] + b[4] * b[11],
            b[1] * b[15] - b[5] * b[13] + b[6] * b[12] - b[7] * b[11],
            -b[2] * b[15] + b[5] * b[14] - b[8] * b[12] + b[9] * b[11],
            b[3] * b[15] - b[6] * b[14] + b[8] * b[13] - b[10] * b[11],
            -b[4] * b[15] + b[7] * b[14] - b[9] * b[13] + b[10] * b[12],
        ]
    )


def null_quadric(r):
    """Evaluates Nullquadric condition for even graded element r.
    Condition is true, if output is 0.

    Parameters
    ----------
    r : cga_object
        object of which to evaluate the null quadric condition

    Returns
    -------
    float

    """
    if r == r.make_even():
        return (r * (~r)).coeff[0]
    else:
        raise GeometryError("Element must be of even grade")
